DEBUG = True

DATABASE_ENGINE = 'sqlite3'
DATABASE_NAME = '/tmp/zums_jd_test_project.db'
INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.admin',

    'zums_jd_test_project'
]
ROOT_URLCONF = 'zums_jd_test_project.urls'

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)
SESSION_ENGINE = "zums.django_backends.session"
AUTHENTICATION_BACKENDS = ["zums.django_backends.auth.ZUMSBackend"]
SECRET_KEY = 'gw2)POSOBX1Z1N4NQ-oh3kv%atu+awg@=@r!jklo@i_v&#o'




import msgpack
from django.contrib.auth.models import User
from zums.zumsd import query

class ZUMSBackend:
    supports_object_permissions = False
    supports_anonymous_user = False
    supports_inactive_user = False

    def update_user_if_required(self, user, user_info):
        save = False
        def check_attr(name, default=""):
            if getattr(user, name) != user_info.get(name, default):
                setattr(user, name, user_info.get(name, default))
                save = True
        check_attr("email")
        check_attr("first_name")
        check_attr("last_name")
        check_attr("is_superuser", False)
        check_attr("is_staff", False)
        if not user.id: save = True
        if save: user.save()

    def authenticate(self, username=None, password=None):
        user_info = query(str("user_authenticate:%s:%s" % (username, password)))
        if user_info:
            user_info = msgpack.loads(user_info)
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                user = User(username=user_info["username"], password='get from zums')
            self.update_user_if_required(user, user_info)
            return user
        return None

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



import zmq, threading, time, logging, msgpack

CONTEXT = zmq.Context()
ZNull = zmq.Message(None)

# multi helpers # {{{
def recv_multi(sock):
    parts = []
    while True:
        parts.append(sock.recv())
        if not sock.getsockopt(zmq.RCVMORE): break
    return parts

def send_multi(sock, parts, reply=None):
    if reply is not None:
        parts[-1] = reply
    for part in parts[:-1]:
        sock.send(part, zmq.SNDMORE)
    sock.send(parts[-1], 0)
# }}}

logger = logging.getLogger("zums.ZReplier")
class NoReply(Exception): pass

# ZReplier # {{{
class ZReplier(threading.Thread):

        def __init__(self, bind):
            super(ZReplier, self).__init__()
            self.shutdown_event = threading.Event()
            self.daemon = True
            self.bind = bind
            self.stats = {}
            self.stats["started_on"] = time.asctime()

        def thread_init(self):
            self.socket = CONTEXT.socket(zmq.XREP)
            try:
                self.socket.bind(self.bind)
            except zmq.ZMQError, e:
                print e, "while binding on", self.bind
                self.shutdown_event.set()
                raise

        def thread_quit(self):
            self.socket.close()

        def reply(self, message):
            if message == "shutdown":
                self.shutdown_event.set()
                logger.warn("shutdown")
                self.increment_stats_counter("shutdown")
                return "shutting down"
            if message == "stats":
                logger.info("stats")
                self.increment_stats_counter("stats")
                return msgpack.dumps(self.stats)
            self.increment_stats_counter("no_reply")
            raise NoReply

        def increment_stats_counter(self, counter_name):
            if counter_name not in self.stats:
                self.stats[counter_name] = 0
            self.stats[counter_name] += 1

        def run(self):
            self.thread_init()

            print self.__class__.__name__, "listening on %s." % self.bind


            while not self.shutdown_event.isSet():
                parts = recv_multi(self.socket)

                self.increment_stats_counter("requests")

                if len(parts) != 3:
                    logger.warn(
                        "Expected 3 parts, got %s: %s" % (len(parts), parts)
                    )
                    send_multi(self.socket, parts, "BAD MESSAGE")
                    continue

                message = parts[2]

                try:
                    send_multi(self.socket, parts, self.reply(message))
                except NoReply:
                    logger.info("NoReply for: %s" % message)
                    send_multi(self.socket, parts, "Unknown command.")
                except Exception, e:
                    logger.exception("Exception %s for: %s" % (e, message))
                    send_multi(self.socket, parts, "exception: %s" % e)

            self.thread_quit()

        def shutdown(self):
            socket = CONTEXT.socket(zmq.REQ)
            socket.connect(self.bind)

            socket.send("shutdown")
            recv_multi(socket)
            socket.close()

        def loop(self):
            self.start()
            try:
                while True:
                    self.shutdown_event.wait(1)
                    if self.shutdown_event.isSet():
                        print "Terminating after remote signal."
                        break
            except KeyboardInterrupt:
                print "Terminating gracefully... "
                self.shutdown()
                self.join()
                print "Terminated."
# }}}

def query_maker(bind):
    socket = CONTEXT.socket(zmq.REQ)
    socket.connect(bind)

    def query(*args, **kw):
        if args and kw:
            cmd = "%s:%s" % ( ":".join(args), msgpack.dumps(kw))
        elif args:
            cmd = ":".join(args)
        elif kw:
            cmd = msgpack.dumps(kw)
        else:
            cmd = ""
        socket.send(cmd)
        return socket.recv()

    return query




# imports and config # {{{
from zreplier import ZReplier, query_maker
import bsddb, time, logging, os, random, hashlib, msgpack, argparse

if hasattr(random, 'SystemRandom'):
    randrange = random.SystemRandom().randrange
else:
    randrange = random.randrange

ZUMS_BIND = "tcp://127.0.0.1:7979"
logger = logging.getLogger("zumsd")
MAX_SESSION_KEY = 18446744073709551616L     # 2 << 63
# }}}

# BDBSessionStore # {{{
class BDBSessionStore:
    def __init__(self, db_file="./sessions.bdb"):
        self.db_file = db_file
        self.db = bsddb.hashopen(db_file)
        print "BDBSessionStore initialized with %s, containing %s sessions." % (
            db_file, len(self.db)
        )

    def get_next_key(self):
        try:
            pid = os.getpid()
        except AttributeError:
            # No getpid() in Jython, for example
            pid = 1
        while 1:
            session_key = hashlib.md5(
                "%s%s%s" % (randrange(0, MAX_SESSION_KEY), pid, time.time())
            ).hexdigest()
            if not self.exists(session_key):
                break
        return session_key

    def create(self):
        session = {}
        sid = self.get_next_key()
        session["sessionid"] = sid
        session["created_on"] = time.time()
        self.db[sid] = msgpack.dumps(session)
        return sid

    def get(self, sid):
        return self.db[sid]

    def set(self, sid, value):
        data = msgpack.loads(value)
        data.pop("sessionid", None) # ignore sessionid, it can not be changed
        self.db[sid] = msgpack.dumps(data)
        return "OK"

    def exists(self, sid):
        return sid in self.db # and checks TODO

    def delete(self, sid):
        if sid in self.db: del self.db[sid]
# }}}

# ZUMSServer # {{{
class ZUMSServer(ZReplier):
    def __init__(self, bind, sessionfile):
        super(ZUMSServer, self).__init__(bind)
        self.sessionfile = sessionfile

    def thread_init(self):
        super(ZUMSServer, self).thread_init()
        self.sessions = BDBSessionStore(self.sessionfile)

    def authenticate(self, line):
        from django.contrib.auth.models import User
        parts = line.split(":", 2)
        if len(parts) < 3: return ""
        try: user = User.objects.get(username=parts[1])
        except User.DoesNotExist: return ""
        if not user.check_password(parts[2]): return ""
        return msgpack.dumps(
            {
                "username": user.username,
                "email": user.email,
                "first_name": user.first_name,
                "last_name": user.last_name,

                "is_superuser": user.is_superuser,
                "is_staff": user.is_staff,
                "is_active": user.is_active,
            }
        )

    def reply(self, line):
        if line == "session_create":
            return self.sessions.create()
        if line.startswith("session_get:"):
            parts = line.split(":")
            return self.sessions.get(parts[1])
        if line.startswith("session_exists:"):
            parts = line.split(":")
            return "yes" if self.sessions.exists(parts[1]) else ""
        if line.startswith("session_set:"):
            parts = line.split(":", 2)
            return self.sessions.set(parts[1], parts[2])
        if line.startswith("session_create:"):
            parts = line.split(":", 2)
            if self.sessions.exists(parts[1]): return "ZUMS.SessionExists"
            return self.sessions.set(parts[1], parts[2])
        if line.startswith("session_delete:"):
            parts = line.split(":")
            self.sessions.delete(parts[1])
            return "OK"
        if line.startswith("user_authenticate:"):
            return self.authenticate(line)
        return super(ZUMSServer, self).reply(line)
# }}}

query = query_maker(bind=ZUMS_BIND)

def main():
    parser = argparse.ArgumentParser(
        description='zmq based session and user manager for web applications.'
    )
    parser.add_argument("--bind", default=ZUMS_BIND)
    parser.add_argument("--sessionfile", default="./sessions.bdb")
    parser.add_argument("--settings", default="zums.zumsd_users.settings")
    arguments = parser.parse_args()
    os.environ["DJANGO_SETTINGS_MODULE"] = arguments.settings
    ZUMSServer(arguments.bind, arguments.sessionfile).loop()

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    try:
        main()
    except Exception:
        logger.exception("Exception in main")
        raise



from django.http import HttpResponse

def session_test(request):
    request.session.setdefault("count", 0)
    request.session["count"] += 1
    return HttpResponse("session: %s" % request.session["count"])



from setuptools import setup, find_packages

setup(
    name = "zums",
    version = "0.1.0",
    url = 'http://github.com/amitu/zums',
    license = 'BSD',
    description = file("README.rst").read(),
    author = 'Amit Upadhyay',
    packages = find_packages('ROOT'),
    package_dir = {'': 'ROOT'},
    install_requires = ['setuptools'],
    entry_points={
      'console_scripts': ['zumsd = zums.zumsd:main']
    }
)



from django.contrib.sessions.backends.base import SessionBase, CreateError
from zums.zumsd import query
import msgpack

class SessionStore(SessionBase):
    """ Implements ZUMS session store. """

    def __init__(self, session_key=None):
        super(SessionStore, self).__init__(session_key)

    def load(self):
        return msgpack.loads(query("session_get:%s" % self.session_key))

    def create(self):
        self.session_key = query("session_create")
        self.modified = True
        self._session_cache = {}

    def exists(self, session_key):
        return bool(query("session_exists:%s" % session_key))

    def save(self, must_create=False):
        if must_create:
            if query(
                "session_create:%s:%s" % (
                    self.session_key,
                    msgpack.dumps(self._get_session(no_load=must_create))
                )
            ) == "ZUMS.SessionExists":
                raise CreateError
        else:
            query(
                "session_set:%s:%s" % (
                    self.session_key,
                    msgpack.dumps(self._get_session(no_load=must_create))
                )
            )

    def delete(self, session_key=None):
        if session_key is None:
            if self._session_key is None:
                return
            session_key = self._session_key
        query("session_delete:%s" % session_key)
